"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
var _eventemitter = _interopRequireDefault(require("../utils/eventemitter.js"));
var _publication = _interopRequireDefault(require("../publication/publication.js"));
var _container = _interopRequireDefault(require("./container.js"));
var _packaging = _interopRequireDefault(require("./packaging.js"));
var _navigation = _interopRequireDefault(require("./navigation.js"));
var _pagelist = _interopRequireDefault(require("./pagelist.js"));
var _spine = _interopRequireDefault(require("./spine.js"));
var _url = require("../utils/url.js");
var _epubcfi = _interopRequireDefault(require("../utils/epubcfi.js"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
const CONTAINER_PATH = "META-INF/container.xml";
const IBOOKS_DISPLAY_OPTIONS_PATH = "META-INF/com.apple.ibooks.display-options.xml";
const INPUT_TYPE = {
  EPUB: "epub",
  OPF: "opf",
  DIRECTORY: "directory"
};
class Epub extends _publication.default {
  constructor(url, options) {
    super();
    if (url) {
      this.opened = this.open(url);
    }
  }

  /**
   * Determine the type of they input passed to open
   * @private
   * @param  {string} input
   * @return {string}  directory | epub | opf
   */
  determineType(input) {
    const ext = (0, _url.extension)(input);
    if (!ext) {
      return INPUT_TYPE.DIRECTORY;
    }
    if (ext === "epub") {
      return INPUT_TYPE.EPUB;
    }
    if (ext === "opf") {
      return INPUT_TYPE.OPF;
    }
  }

  /**
   * Open the epub container
   * @private
   * @param  {string} url
   * @return {string} packagePath
   */
  async loadContainer(url) {
    const xml = await this.load(url);
    this.container = new _container.default(xml);
    return this.container.packagePath;
  }

  /**
   * Open the Open Packaging Format Xml
   * @private
   * @param  {string} url
   * @return {Promise}
   */
  async loadPackaging(url) {
    const xml = await this.load(url);
    this.packaging = new _packaging.default(xml);
    return this.packaging;
  }

  /**
   * Load Navigation and PageList from package
   * @private
   * @param {document} opf XML Document
   */
  async loadNavigation(opf) {
    let navPath = opf.navPath || opf.ncxPath;
    if (!navPath) {
      return {
        toc: undefined,
        landmarks: undefined,
        pageList: undefined,
        locations: undefined
      };
    }
    const xml = await this.load(navPath, "xml");
    const navigation = new _navigation.default(xml, this.resolve(navPath));
    const pagelist = new _pagelist.default(xml);
    return {
      toc: navigation.toc,
      landmarks: navigation.landmarks,
      pageList: pagelist.pages,
      locations: pagelist.locations
    };
  }
  async loadDisplayOptions(packaging) {
    let displayOptionsXml;
    if (packaging.metadata.layout === "") {
      displayOptionsXml = await this.load(IBOOKS_DISPLAY_OPTIONS_PATH).catch(err => {
        return undefined;
      });
    }
    return displayOptionsXml;
  }
  loadSections(packaging) {
    let spine = new _spine.default(packaging);
    return {
      readingOrder: spine.readingOrder,
      unordered: spine.resources
    };
  }
  loadResources(packaging) {
    let resources = [];
    for (const r in packaging.manifest) {
      const resource = packaging.manifest[r];
      if (resource.type !== "application/xhtml+xml" && resource.type !== "text/html") {
        resource.url = resource.href;
        resources.push(resource);
      }
    }
    return resources;
  }
  loadMetadata(packaging) {
    return packaging.metadata;
  }

  /**
   * Unpack the contents of the Epub
   * @private
   * @param {document} packageXml XML Document
   */
  async unpack(packaging) {
    this.package = packaging;
    this.metadata = this.loadMetadata(packaging);
    const resources = this.loadResources(packaging);
    const {
      readingOrder,
      unordered
    } = this.loadSections(packaging);
    this.spine = readingOrder;
    this.resources = [...unordered, ...resources];
    const {
      toc,
      landmarks,
      pageList,
      locations
    } = await this.loadNavigation(packaging);
    this.navigation = toc;
    this.landmarks = landmarks;
    this.pagelist = pageList;
    this.locations = locations;
    this.displayOptions = await this.loadDisplayOptions(packaging);
    return this;
  }
  async open(url, what) {
    const type = what || this.determineType(url);
    let packaging;
    this.url = url;
    if (type === INPUT_TYPE.EPUB) {
      throw new Error("Epub must be unarchived");
    }
    if (type === INPUT_TYPE.DIRECTORY) {
      const container = await this.loadContainer(CONTAINER_PATH);
      packaging = await this.loadPackaging(container.packagePath);
    } else {
      packaging = await this.loadPackaging(url);
    }
    return this.unpack(packaging);
  }
  get spine() {
    return this.sections;
  }
  set spine(items) {
    return this.sections = items;
  }
  get toc() {
    return this.navigation;
  }
  set toc(items) {
    return this.navigation = items;
  }
  toJSON() {}

  /**
   * Destroy the Epub and all associated objects
   */
  destroy() {}
}
(0, _eventemitter.default)(Epub.prototype);
var _default = Epub;
exports.default = _default;