"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
var _core = require("../utils/core.js");
var _epubcfi = _interopRequireDefault(require("../utils/epubcfi.js"));
var _eventemitter = _interopRequireDefault(require("../utils/eventemitter.js"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
/**
 * Locators
 * @param {object} [manifest]
 */
class Locators {
  constructor(publication) {
    if (publication) {
      this.unpack(publication);
    }
  }
  unpack(publication) {
    if (publication.locations) {
      this.unpackLocations(publication.locations);
    }
    if (publication.pagelist) {
      this.unpackPages(publication.pagelist);
    }
  }
  unpackLocations(locations) {
    if (!locations) {
      return;
    }
    this.locations = [];
    for (const [key, location] of locations) {
      if (_epubcfi.default.prototype.isCfiString(location)) {
        this.locations.push(location);
      } else if (location.cfi) {
        this.locations.push(location.cfi);
      }
    }
    this.totalLocations = this.locations.length - 1;
  }
  unpackPages(pages) {
    if (!pages) {
      return;
    }
    this.pages = pages;
    this.pageLocations = [];
    this.firstPage = parseInt(this.pages[0]);
    this.lastPage = parseInt(this.pages[this.pages.length - 1]);
    this.totalPages = this.lastPage - this.firstPage;
    pages.forEach(item => {
      if (item.cfi) {
        this.pageLocations.push(item.cfi);
      }
    });
  }

  /**
   * Get a location from an EpubCFI
   * @param {EpubCFI} cfi
   * @return {number}
   */
  locationFromCfi(cfi) {
    let loc;
    if (_epubcfi.default.prototype.isCfiString(cfi)) {
      cfi = new _epubcfi.default(cfi);
    }
    // Check if the location has not been set yet
    if (this.locations.length === 0) {
      return -1;
    }
    loc = (0, _core.locationOf)(cfi, this.locations, _epubcfi.default.prototype.compare);
    if (loc > this.totalLocations) {
      return this.totalLocations;
    }
    return loc;
  }

  /**
   * Get a percentage position in locations from an EpubCFI
   * @param {EpubCFI} cfi
   * @return {number}
   */
  percentageFromCfi(cfi) {
    if (!this.locations || this.locations.length === 0) {
      return null;
    }
    // Find closest cfi
    let loc = this.locationFromCfi(cfi);
    // Get percentage in total
    return this.percentageFromLocation(loc);
  }

  /**
   * Get a percentage position from a location index
   * @param {number} location
   * @return {number}
   */
  percentageFromLocation(loc) {
    if (!loc || !this.totalLocations) {
      return 0;
    }
    return loc / this.totalLocations;
  }

  /**
   * Get an EpubCFI from location index
   * @param {number} loc
   * @return {EpubCFI} cfi
   */
  cfiFromLocation(loc) {
    var cfi = -1;
    // check that pg is an int
    if (typeof loc != "number") {
      loc = parseInt(loc);
    }
    if (loc >= 0 && loc < this.locations.length) {
      cfi = this.locations[loc];
    }
    return cfi;
  }

  /**
   * Get an EpubCFI from location percentage
   * @param {number} percentage
   * @return {EpubCFI} cfi
   */
  cfiFromPercentage(percentage) {
    let loc;
    if (percentage > 1) {
      console.warn("Normalize cfiFromPercentage value to between 0 - 1");
    }

    // Make sure 1 goes to very end
    if (percentage >= 1) {
      let cfi = new _epubcfi.default(this.locations[this.totalLocations]);
      cfi.collapse();
      return cfi.toString();
    }
    loc = Math.ceil(this.totalLocations * percentage);
    return this.cfiFromLocation(loc);
  }

  /**
   * Get a PageList result from a EpubCFI
   * @param  {string} cfi EpubCFI String
   * @return {string} page
   */
  pageFromCfi(cfi) {
    var pg = -1;

    // Check if the pageList has not been set yet
    if (!this.pageLocations || this.pageLocations.length === 0) {
      return -1;
    }

    // check if the cfi is in the location list
    var index = indexOfSorted(cfi, this.pageLocations, _epubcfi.default.prototype.compare);
    if (index != -1) {
      pg = this.pages[index];
    } else {
      // Otherwise add it to the list of locations
      // Insert it in the correct position in the locations page
      index = (0, _core.locationOf)(cfi, this.pageLocations, _epubcfi.default.prototype.compare);
      // Get the page at the location just before the new one, or return the first
      pg = index - 1 >= 0 ? this.pages[index - 1] : this.pages[0];
      if (pg !== undefined) {
        // Add the new page in so that the locations and page array match up
        //this.pages.splice(index, 0, pg);
      } else {
        pg = -1;
      }
    }
    return pg;
  }

  /**
   * Get an EpubCFI from a Page List Item
   * @param  {string} pg
   * @return {string} cfi
   */
  cfiFromPage(pg) {
    var cfi = -1;
    // check that pg is an int
    if (typeof pg != "number") {
      pg = parseInt(pg);
    }

    // check if the cfi is in the page list
    // Pages could be unsorted.
    var index = this.pages.indexOf(pg);
    if (index != -1) {
      cfi = this.pageLocations[index];
    }
    // TODO: handle pages not in the list
    return cfi;
  }

  /**
   * Get a Page from Book percentage
   * @param  {number} percent
   * @return {string} page
   */
  pageFromPercentage(percent) {
    var pg = Math.round(this.totalPages * percent);
    return pg;
  }

  /**
   * Returns a value between 0 - 1 corresponding to the location of a page
   * @param  {int} pg the page
   * @return {number} percentage
   */
  percentageFromPage(pg) {
    var percentage = (pg - this.firstPage) / this.totalPages;
    return Math.round(percentage * 1000) / 1000;
  }

  /**
   * Returns a value between 0 - 1 corresponding to the location of a cfi
   * @param  {string} cfi EpubCFI String
   * @return {number} percentage
   */
  percentagePageFromCfi(cfi) {
    var pg = this.pageFromCfi(cfi);
    var percentage = this.percentageFromPage(pg);
    return percentage;
  }
  destroy() {}
}
(0, _eventemitter.default)(Locators.prototype);
var _default = Locators;
exports.default = _default;